#!/usr/bin/perl -w
#
# STRedundancy.pl v1.0
# Last modified July 27, 2005 13:29
# (c) Olaf R.P. Bininda-Emonds
#
# Input: Nexus file containing data matrix and list of taxa to be deleted listed in a taxset
#
# Output: A tab-delimited file (*_STRredundancy.txt) listing characters that would become redundant
#		  (i.e., duplicate others or are uninformative) if the taxa were to be deleted. Searches for
#		  duplicate characters will be restricxted to charsets of these are optionally given.
#
# Usage: STRedundancy.pl -d<filename> [-h]
#	options: -d<filename> = name of nexus file containing data matrix and taxa to be deleted in a taxset statement
#			 -h = print this message and quit

use strict;

# Set default parameters
	# I/O parameters
		my ($nexusFile, $dataSource, $outFile);
		
	# Redundancy variables
		my %character;
		my %duplChar;
		my @uninfList;

	# Miscellaneous parameters
		my $verbose = 0;
		my $debug = 0;

# Get command line input
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] =~ /-d(.*)/)
			{
			$nexusFile = $1;
			($dataSource = $nexusFile) =~ s/\.\w+$//;
			$outFile = $dataSource . "_STRredundancy.txt";
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$verbose = $debug = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: STRedundancy.pl -d<filename> [-h]\n";
			print "Options: -d<filename> = name of nexus file containing data matrix and taxa to be deleted in a taxset statement\n";
			print "         -h = print this message and quit\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: STRedundancy.pl -d<filename> [-h]\n";
			exit(1); 
			}
		}

# Safeties
	die "ERROR: Must provide name of nexus file containing data matrix and taxa to be deleted in a taxset statement\n" if (not defined $nexusFile);

# Read in and process nexus file
	print "Reading in data set and sets from file $nexusFile ...\n" if ($verbose);
	setLineBreak($nexusFile);
	open (DATA, "<$nexusFile") or die "Cannot open nexus file containing data set, $nexusFile\n";
		my $matrixRead = 0;
		my (@taxonList, %charStates);
		
		my (@charSets, %charSetStart, %charSetStop, %charSetID);
		
		my $taxSetRead;
		my %delTaxa;
		my $delCount = 0;
		
		while (<DATA>)
			{
			chomp;
			my $lineRead = $_;
				$lineRead =~ s/^\s+//;
				$lineRead =~ s/\s+$//;
			next unless ($lineRead);
			
			# Read in data matrix
				if ($lineRead =~ /^matrix/i)
					{
					$matrixRead = 1;
					next;
					}
					
				if ($matrixRead)
					{
					if ($lineRead =~ /;/)
						{
						$matrixRead = 0;
						next;
						}
					else
						{
						my ($species, $characters) = split(/\s+/, $lineRead);
						push @taxonList, $species;
							$delTaxa{$species} = 0;
						$charStates{$species} = $characters;
						}
					}
					
			# Read in charsets
				if ($lineRead =~ /^charset/i)
					{
					$lineRead =~ s/^charset//i;
					$lineRead =~ s/;$//i;
					my ($name, $bounds) = split(/\s?\=\s?/, $lineRead);
					push @charSets, $name;
						my ($start, $stop) = split(/\s?\-\s?/, $bounds);
							$stop = $start if (not defined $stop);	# Safety for single character charsets
						$charSetStart{$name} = $start;
						$charSetStop{$name} = $stop;
						$charSetID{$_} = $name for($start..$stop);
					}

			# Read in taxsets
				if ($lineRead =~ /^taxset/i and $delCount == 0)
					{
					$taxSetRead = 1;
					$lineRead =~ s/^taxset//i;
					if ($lineRead =~ /;/)
						{
						$taxSetRead = 0;
						$lineRead =~ s/;$//i;
						}
					my ($name, $taxaList) = split(/\s?\=\s?/, $lineRead);
					foreach my $taxon (split(/\s+/, $taxaList))
						{
						$delTaxa{$taxon} = 1;
						$delCount++;
						}
					next;
					}
					
				if ($taxSetRead)
					{
					if ($lineRead =~ /;/)
						{
						$taxSetRead = 0;
						$lineRead =~ s/;$//i;
						}
					foreach my $taxon (split(/\s+/, $lineRead))
						{
						$delTaxa{$taxon} = 1;
						$delCount++;
						}
					}

				}
	close DATA;

die "ERROR: Must specify list of taxa to be deleted in a taxset statement.\n" if (not $delCount);

my $charNum = length($charStates{$taxonList[0]});

if ($debug)
	{
	printf "Data matrix (%s taxa; %s characters):\n", scalar(@taxonList), length($charStates{$taxonList[0]});
		print "\t$_\t$charStates{$_}\n" foreach (@taxonList);
	printf "\nCharacter sets (%s):\n", scalar(@charSets);
		print "\t$_\t$charSetStart{$_}\t$charSetStop{$_}\n" foreach (@charSets);
	print "\nTaxon set ($delCount):\n";
		foreach my $taxon (@taxonList)
			{
			print "\t$taxon" if ($delTaxa{$taxon} == 1);
			}
		print "\n";
	}
	
# Process data matrix to find identical and, optionally, uninformative characters
	# Transpose matrix to get character lists
		for (my $char = 1; $char <= $charNum; $char++)
			{
			foreach my $taxon (@taxonList)
				{
				next if $delTaxa{$taxon};
				$character{$char} .= substr($charStates{$taxon}, $char - 1, 1);
				}
			}
		if ($debug)
			{
			for (my $char = 1; $char <= $charNum; $char++)
				{
				print "$char\t$character{$char}\n";
				}
			}
	
	# Find identical characters
		print "Checking for identical character ...\n" if ($verbose);
		if (@charSets)	# Do for each individual character set
			{
			charMatcher($charSetStart{$_}, $charSetStop{$_}) foreach (@charSets);
			}
		else	# Do over global character set
			{
			charMatcher(1, $charNum);
			}
				
	# Print out results
		open (OUT, ">$outFile") or die "Cannot open results file, $outFile\n";
		print OUT "Character";
			print OUT "\tCharset" if (@charSets);
		print OUT "\tIdentical with\tUninformative?\n";
		
		for (my $char = 1; $char <= $charNum; $char++)
			{
			print OUT "$char";
			print OUT "\t$charSetID{$char}" if (@charSets);
			if (defined @ { $duplChar{$char} })
				{
				printf OUT "\t%s", join(', ', @ { $duplChar{$char} });
				}
			else
				{
				print OUT "\tnone";
				}
			
			# Check for uninformative characters
				my $count0 = ($character{$char} =~ tr/0//);
				my $count1 = ($character{$char} =~ tr/1//);
				if ($count0 == 1 or $count1 == 1)
					{
					print OUT "\tyes";
					push @uninfList, $char;
					}
			print OUT "\n";
			}

		if (@uninfList)
			{
			@uninfList = duplStrip(@uninfList);
			@uninfList = sort @uninfList;
			printf OUT "\ncharset STR = %s;\n", join(' ', @uninfList);
			}
		close OUT;

exit(0);

### Subroutines used in the program

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
	
sub charMatcher
	{
	my $startChar = shift;
	my $stopChar = shift;
	
	for (my $refChar = $startChar; $refChar <= $stopChar; $refChar++)
		{
		print "Reference character: $refChar\n" if ($debug);
		for (my $compChar = $startChar; $compChar <= $stopChar; $compChar++)
			{
			next if ($refChar == $compChar);
			print "\tComparison character: $compChar\n" if ($debug);
			if ($character{$refChar} eq $character{$compChar})
				{
				push @ { $duplChar{$compChar} }, $refChar;
				push @uninfList, $compChar if ($compChar > $refChar);
				print "\t\tidentical\n" if ($debug);
				}
			}
		}
	return;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

# Version history
#
#	v1.0 (July 27, 2005)
#		- initial release
